/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2012 University of Bonn. All rights reserved.
 * Please see the COPYING file or "Copyright notice" in builder.cpp for details.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>. 
 */

/*
 * PairPotential_Harmonic.cpp
 *
 *  Created on: Sep 26, 2012
 *      Author: heber
 */


// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "CodePatterns/MemDebug.hpp"

#include "PairPotential_Harmonic.hpp"

#include "CodePatterns/Assert.hpp"

#include "Potentials/helpers.hpp"

PairPotential_Harmonic::PairPotential_Harmonic() :
  params(parameters_t(MAXPARAMS, 0.))
{}

PairPotential_Harmonic::PairPotential_Harmonic(
    const double _spring_constant,
    const double _equilibrium_distance,
    const double _energy_offset) :
      params(parameters_t(MAXPARAMS, 0.))
{
  params[spring_constant] = _spring_constant;
  params[equilibrium_distance] = _equilibrium_distance;
  params[energy_offset] = _energy_offset;
}

PairPotential_Harmonic::results_t
PairPotential_Harmonic::operator()(
    const arguments_t &arguments
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_Harmonic::operator() - requires exactly one argument.");
  const argument_t &r_ij = arguments[0];
  const result_t result =
      params[spring_constant]
             * Helpers::pow( r_ij.distance - params[equilibrium_distance], 2 )
            + params[energy_offset];
  return std::vector<result_t>(1, result);
}

PairPotential_Harmonic::derivative_components_t
PairPotential_Harmonic::derivative(
    const arguments_t &arguments
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_Harmonic::operator() - requires exactly one argument.");
  derivative_components_t result;
  const argument_t &r_ij = arguments[0];
  result.push_back( 2. * params[spring_constant] * ( r_ij.distance - params[equilibrium_distance]) );
  ASSERT( result.size() == 1,
      "PairPotential_Harmonic::operator() - we did not create exactly one component.");
  return result;
}

PairPotential_Harmonic::results_t
PairPotential_Harmonic::parameter_derivative(
    const arguments_t &arguments,
    const size_t index
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_Harmonic::parameter_derivative() - requires exactly one argument.");
  const argument_t &r_ij = arguments[0];
  switch (index) {
    case spring_constant:
    {
      const result_t result =
                 Helpers::pow( r_ij.distance - params[equilibrium_distance], 2 );
      return std::vector<result_t>(1, result);
      break;
    }
    case equilibrium_distance:
    {
      const result_t result =
          -2. * params[spring_constant]
                 * ( r_ij.distance - params[equilibrium_distance]);
      return std::vector<result_t>(1, result);
      break;
    }
    case energy_offset:
    {
      const result_t result = +1.;
      return std::vector<result_t>(1, result);
      break;
    }
    default:
      break;
  }

  return PairPotential_Harmonic::results_t(1, 0.);
}

